/* Copyright Cloud Linux Inc 2010-2011 All Rights Reserved
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * This is the hostinglimits module for apache 2.X
 * author Igor Seletskiy <iseletsk@cloudlinux.com>
 * author Alexey Berezhok <alexey.berezhok@cloudlinux.com>
 *
 */
#define LVE_DEPRICATED

#include "apr_general.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "ap_config.h"
#include "util_filter.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "http_core.h"
#include <unistd.h>
#include "unixd.h"
#include "mpm_common.h"

#include <lve/lve-ctl.h>

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
                                                                                                                                          
#include <pthread.h>                                                                                                                                         
#include <unistd.h>                                                                                                                                          
#include <stdlib.h>                                                                                                                                          
#include <sched.h>                                                                                                                                           
#include <linux/unistd.h>                                                                                                                                    
#include <sys/syscall.h>                                                                                                                                     
#include <errno.h> 

#define PREFIX "mod_hostinglimits:"

#define MOD_HOSTINGLIMITS_NEED_DEBUG 4
#define MOD_HOSTINGLIMITS_UNNEED_DEBUG 1

#define CHECK_DEBUG if(need_debug==MOD_HOSTINGLIMITS_NEED_DEBUG)

module AP_MODULE_DECLARE_DATA hostinglimits_module;
static apr_threadkey_t *key, *debug_key;

// configuration data
typedef struct hostinglimits_module_cfg {
  unsigned int skip;
  uint32_t lve_id;
  apr_array_header_t *allowed_handlers;
  const char *lve_user;
  apr_array_header_t *debug_sites;
  uid_t ruid_uid;
  uid_t itk_uid;
} hostinglimits_module_cfg;

static int lve_available = 0;
struct liblve *lve = NULL;

int get_need_debug(apr_array_header_t *list_of_sites, request_rec *r){
  int num_names;
  char **names_ptr;
  if (!r->hostname) return MOD_HOSTINGLIMITS_UNNEED_DEBUG;
  if(list_of_sites){
    names_ptr = (char **)list_of_sites->elts;
    num_names = list_of_sites->nelts;
    for (; num_names; ++names_ptr, --num_names) {
          if (strstr(r->hostname, *names_ptr)) {
            return MOD_HOSTINGLIMITS_NEED_DEBUG;
          }
    }

  }
  return MOD_HOSTINGLIMITS_UNNEED_DEBUG;
}

pid_t gettid( void )                                                                                                                                         
{                                                                                                                                                            
    return syscall( __NR_gettid );                                                                                                                           
}                                                                                                                                                            
       

/**
 * This function returns directory config structure for current request
 */
static hostinglimits_module_cfg *hostinglimits_module_dconfig(const request_rec *r) {
  return (hostinglimits_module_cfg *) ap_get_module_config(r->per_dir_config,
                                                           &hostinglimits_module);
}


static void *hostinglimits_module_merge_config(apr_pool_t* p, void* BASE, void* ADD) {
  hostinglimits_module_cfg* base = BASE;
  hostinglimits_module_cfg* add = ADD;
  hostinglimits_module_cfg* cfg = (hostinglimits_module_cfg *)
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  cfg->skip = (add->skip)?add->skip:base->skip;
  cfg->lve_id = (add->lve_id)?add->lve_id:base->lve_id;
  cfg->allowed_handlers = (add->allowed_handlers)?add->allowed_handlers:
    base->allowed_handlers;
  cfg->lve_user = (add->lve_user)?add->lve_user:base->lve_user;
  cfg->debug_sites = (add->debug_sites)?add->debug_sites:
    base->debug_sites;
  cfg->itk_uid = (add->itk_uid)?add->itk_uid:base->itk_uid;
  cfg->ruid_uid = (add->ruid_uid)?add->ruid_uid:base->ruid_uid;
  return cfg;
}

static void *hostinglimits_module_create_dir_config(apr_pool_t *p,
                                                    char *dirspec) {
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *)
    apr_pcalloc(p, sizeof(hostinglimits_module_cfg));
  if(!cfg) {
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, NULL,
                   PREFIX" not enough memory");
    return NULL;
  }
  cfg->skip  = 0;
  cfg->lve_id = 0;
  cfg->allowed_handlers = NULL;
  cfg->lve_user = NULL;
  cfg->debug_sites = NULL;
  cfg->itk_uid = -1;
  cfg->ruid_uid = -1;
  return (void *) cfg;
}


static const char *set_lve_id(cmd_parms *cmd, void *mcfg,
                              const char *lve_id)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_id) {
    cfg->lve_id = (uint32_t)apr_atoi64(lve_id);
  }
  return NULL;
}

static const char *set_lve_user(cmd_parms *cmd, void *mcfg,
                              const char *lve_user)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (lve_user) {
    cfg->lve_user = lve_user;
  }
  return NULL;
}


static const char *set_handlers(cmd_parms *cmd, void *mcfg, const char *arg)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }
  if (!cfg->allowed_handlers) {
    cfg->allowed_handlers = apr_array_make(cmd->pool, 2, sizeof(char *));
  }
  *(const char **)apr_array_push(cfg->allowed_handlers) = arg;
  return NULL;
}

static const char *set_assign_user_id(cmd_parms *cmd, void *mcfg,
                                      const char *user_name, const char *group_name)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (user_name) {
    cfg->itk_uid = ap_uname2id(user_name);
  }
  return NULL;
}

static const char *set_uidgid(cmd_parms *cmd, void *mcfg,
                                      const char *user_name, const char *group_name)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  if (user_name) {
    cfg->ruid_uid = ap_uname2id(user_name);
  }
  return NULL;
}


static const char *set_debug(cmd_parms *cmd, void *mcfg, const char *arg)
{
  hostinglimits_module_cfg *cfg = (hostinglimits_module_cfg *) mcfg;
  const char *err = ap_check_cmd_context(cmd, NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
  if (err != NULL) {
    return err;
  }
  if (!cfg->debug_sites) {
    cfg->debug_sites = apr_array_make(cmd->pool, 2, sizeof(char *));
  }
  *(const char **)apr_array_push(cfg->debug_sites) = arg;
  return NULL;
}

// describe used directives
static command_rec hostinglimits_module_directives[] = {
  AP_INIT_FLAG("SkipErrors", ap_set_flag_slot,  (void *)
               APR_OFFSETOF(hostinglimits_module_cfg, skip),
               RSRC_CONF, "Allow apache to continue even if LVE is unavalable"),
  AP_INIT_TAKE1("LVEId", set_lve_id, NULL, ACCESS_CONF | RSRC_CONF, "LVE Id"),
  AP_INIT_TAKE1("LVEUser", set_lve_user, NULL, ACCESS_CONF | RSRC_CONF, "LVE User"),
  AP_INIT_ITERATE("AllowedHandlers", set_handlers, NULL, RSRC_CONF,
                  "A list of handlers which will be put in LVE"),
  AP_INIT_TAKE2("AssignUserID", set_assign_user_id, NULL, RSRC_CONF|ACCESS_CONF,
                "Tie a virtual host to a specific child process."),
  AP_INIT_TAKE2 ("RUidGid", set_uidgid, NULL, RSRC_CONF | ACCESS_CONF,
                 "Minimal uid or gid file/dir, else set[ug]id to default (User,Group)"),
  AP_INIT_ITERATE("LVESitesDebug", set_debug, NULL, RSRC_CONF,
                   "A list of sites which should be debugged"),
  {NULL}
};

static int init_post_config(apr_pool_t *pconf, apr_pool_t *plog,
                        apr_pool_t *ptemp, server_rec *s) {
  if (lve_available) {
      ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                   PREFIX"LVE mechanism enabled");
  } else {
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                   PREFIX"LVE mechanism disabled, LVE is unavailable");
  }

  return OK;
}

static int match_handler(apr_array_header_t *allowed_handlers, const char *handler) {
  int num_names;
  char **names_ptr;
  int accept = 0;
  if (allowed_handlers) {
      names_ptr = (char **)allowed_handlers->elts;
      num_names = allowed_handlers->nelts;
    } else {
      return 0;
    }
    // match handle
    for (; num_names; ++names_ptr, --num_names) {
      if (!strcmp(handler, *names_ptr)) {
        accept=1;
        break;
      }
    }
    return accept;
}

/**
 * Where are main functionality in this function.
 * This funcion performs after reading request headers
 * and before other proceccing.
 *
 *
 */
static int enter_lve_handler(request_rec *r) {
  // get directory config info
  uint32_t p_cookie = 0;
  apr_status_t st;
  int need_debug = MOD_HOSTINGLIMITS_UNNEED_DEBUG;
  
  hostinglimits_module_cfg *cfg = hostinglimits_module_dconfig(r);
  
  if (!lve_available || !cfg) {
    return DECLINED;
  }
  
  apr_threadkey_private_get((void *)&need_debug, debug_key);
  if(!need_debug){
    need_debug = get_need_debug(cfg->debug_sites, r);
    st = apr_threadkey_private_set(NULL, debug_key);
    apr_threadkey_private_set((void *)need_debug, debug_key);
  }
  
  /* decline if cookie is set ( p_cookie != 0) */
  st=apr_threadkey_private_get((void *)&p_cookie, key);
  CHECK_DEBUG{
     ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
                  PREFIX"[DEBUG] STEP#0001 PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d POOL %d",
                  p_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool?r->pool:0);
  }
  if(st != APR_SUCCESS || (p_cookie!=0)) {
     return DECLINED;
  }

  //lets find handlers for which HostingLimits should run
  if (!match_handler(cfg->allowed_handlers, r->handler)) {
    return DECLINED;
  }

  //enter into lve
  uid_t    uid = -1;
  //1) check if set LVEId
  int lve_id = cfg->lve_id;
  //2) if LVEId is unset and set LVEUser, get id here
  if (!lve_id && cfg->lve_user){
     //if LVEUser set on, read it
     lve_id = (int) ap_uname2id(cfg->lve_user);
  }

  //Get uid from suexec
  ap_unix_identity_t * ugid = ap_run_get_suexec_identity(r);
  if (ugid == NULL) {
      uid = -1;
  } else {
      uid = ugid->uid;
  }
  //if lve_id still less then 1, it is still unset 
  //itk-support and ruid2 support
  if (lve_id < 1) {
      //3) get uid from suexec
      lve_id = uid;
      
      //4) still less then need, than check ruid2
      //check for ruid2 
      if (lve_id < 1){
        if (cfg->ruid_uid!=-1){
          lve_id = cfg->ruid_uid;
        }
      }
      
      //5) still less then need, than check itk 
      //check for itk, if lve_id still less then 0
      if (lve_id < 1){
        if (cfg->itk_uid!=-1){
          lve_id = cfg->itk_uid;
        }
      }
  }
 
  //Priority steps                                                                                                                                         
  // 1) LVEId                                                                                                                                              
  // 2) LVEUser                                                                                                                                            
  // 3) SuexecUserGroup                                                                                                                                    
  // 4) RUidGid                                                                                                                                            
  // 5) AssignUserID
  CHECK_DEBUG{
          ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
                       PREFIX"[DEBUG] STEP#0002 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d, POOL %d",
                       lve_id, p_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool?r->pool:0);
  }
  if (lve_id > 0) {
    int rc = lve_enter(lve, lve_id, -1, -1, &p_cookie);
    int keep_errno = errno;
    CHECK_DEBUG{
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
        PREFIX"[DEBUG] STEP#0003 LVE(%d) PCOOKIE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno %d POOL %d",
        lve_id, p_cookie, r->handler, r->hostname, r->uri, gettid(), errno, r->pool?r->pool:0);
    }
    if(rc) {
      if (keep_errno == EPERM) { //if already inside LVE
        ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                     PREFIX"Already inside LVE: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno (%d)",
                     lve_id, r->handler, r->hostname, r->uri, gettid(), keep_errno);
        return DECLINED;
      }
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Error on LVE enter: LVE(%d) HANDLER(%s) HOSTNAME(%s) URL(%s) TID(%d) errno (%d)",
                   lve_id, r->handler, r->hostname, r->uri, gettid(), keep_errno);
      return HTTP_SERVICE_UNAVAILABLE;
    }
    st = apr_threadkey_private_set((void *)p_cookie, key);
    
    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Unable to set LVE cookie %d", st);
    }
  }

  return DECLINED;
}

static int leave_lve_handler(request_rec *r) {
  uint32_t p_cookie = 0;
  apr_status_t st;

  if (!lve_available) {
    return DECLINED;
  }
  
  int need_debug;
  apr_threadkey_private_get((void *)&need_debug, debug_key);
  if(need_debug!=MOD_HOSTINGLIMITS_NEED_DEBUG){
    need_debug = MOD_HOSTINGLIMITS_UNNEED_DEBUG;
  }
  apr_threadkey_private_set(NULL, debug_key);

  st=apr_threadkey_private_get((void *)&p_cookie, key);
  CHECK_DEBUG{
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
                     PREFIX"[DEBUG] STEP#0004 PCOOKIE(%d) TID(%d) POOL %d",
                     p_cookie, gettid(), r->pool?r->pool:0);
  }
  
  /* if cookie is not initialized or if it is not 0 -- don't try to leave */
  if (st != APR_SUCCESS || p_cookie == 0 ){
    return DECLINED;
  }

  int rc = lve_leave(lve, &p_cookie);
  if(!rc) {
    /* reset cookie to 0 */
    st = apr_threadkey_private_set(NULL, key);

    if (st != APR_SUCCESS){
      ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
                   PREFIX"Unable to reset LVE cookie %d", st);
    }
    return DECLINED;
  }
  ap_log_error(APLOG_MARK, APLOG_ERR, OK, r->server,
               PREFIX"Error exiting %d", rc);
  return DECLINED;
}


static apr_status_t destroy_key_pool(void * dummy)
{
  apr_threadkey_private_delete(key);
  apr_threadkey_private_delete(debug_key);
  return APR_SUCCESS;
}


static void hostinglimits_child_init(apr_pool_t *p, server_rec *s){
   apr_pool_cleanup_register (p, 0, destroy_key_pool, destroy_key_pool);
   apr_threadkey_private_create(&key, NULL, p);
   apr_threadkey_private_create(&debug_key, NULL, p);
}

// function for hook register
static void hostinglimits_module_register_hooks(apr_pool_t *p) {
  int rc;
  errno = 0;
  if (lve == NULL) {
    lve = apr_palloc(p, lve_instance_init(NULL));
    rc = lve_instance_init(lve);
    if (!rc && !errno) {
      lve_available = 1;
      // register clean up LVE
      apr_pool_cleanup_register(p, lve, (void*)lve_instance_destroy, apr_pool_cleanup_null);
    } else {
      lve_available = 0;
    }
  }

  static const char * const aszPre[] = { "mod_include.c", "mod_php.c", "mod_cgi.c", NULL };
  
  ap_hook_post_config(init_post_config, NULL, NULL, APR_HOOK_MIDDLE);
  // hook executes before request prforming
  ap_hook_handler(enter_lve_handler, NULL, aszPre, APR_HOOK_REALLY_FIRST);
  ap_hook_log_transaction(leave_lve_handler, NULL, NULL, APR_HOOK_LAST);
  ap_hook_child_init(hostinglimits_child_init,NULL,NULL,APR_HOOK_MIDDLE);

}

/**
 * Describing structure of Apache module
 */
module AP_MODULE_DECLARE_DATA hostinglimits_module = {
  STANDARD20_MODULE_STUFF,
  hostinglimits_module_create_dir_config,   /* create directory config */
  hostinglimits_module_merge_config,        /* merging directory config */
  NULL,                                     /* create server config */
  NULL,                                     /* merging server config */
  hostinglimits_module_directives,          /* mapping configuration directives */
  hostinglimits_module_register_hooks       /* registering hooks */
};
